import { system, world, ItemStack, BlockPermutation, MinecraftDimensionTypes } from "@minecraft/server"
import * as debug from "../debug"
import * as util from "../util"

const boomerangList = [
	{ id: "ljw_br:boomerang_wooden" },
	{ id: "ljw_br:boomerang_wooden" },
	{ id: "ljw_br:boomerang_stone" },
	{ id: "ljw_br:boomerang_iron" },
	{ id: "ljw_br:boomerang_golden" },
	{ id: "ljw_br:boomerang_diamond" },
	{ id: "ljw_br:boomerang_netherite" },
	{ id: "ljw_br:boomerang_crimson" },
	{ id: "ljw_br:boomerang_warped" },
	{ id: "ljw_br:boomerang_gilded" },
	{ id: "ljw_br:boomerang_frost" },
	{ id: "ljw_br:boomerang_magma" },
	{ id: "ljw_br:boomerang_magma_charged" },
	{ id: "ljw_br:boomerang_hydrorang" },
	{ id: "ljw_br:boomerang_minerang" },
	{ id: "ljw_br:boomerang_cutterang" },
	{ id: "ljw_br:boomerang_enderang" },
	{ id: "ljw_br:boomerang_echorang" },
	{ id: "ljw_br:boomerang_enigmarang" },
	{ id: "ljw_br:boomerang_slime" },
	{ id: "ljw_br:boomerang_tnt" },
	{ id: "ljw_br:boomerang_gatherang" },
	{ id: "ljw_br:boomerang_wolfwind" },
	{ id: "ljw_br:boomerang_turborang" },
	{ id: "ljw_br:boomerang_void" },
	{ id: "ljw_br:info_book" },
	{ id: "ljw_br:boomerang_breezyrang" },
	{ id: "ljw_br:boomerang_raptorang" },
	{ id: "ljw_br:boomerang_heavenlyrang" },
	{ id: "ljw_br:boomerang_jellyrang" },
	{ id: "ljw_br:boomerang_torcharang" },
	{ id: "ljw_br:boomerang_hopperang" },
	{ id: "ljw_br:boomerang_collectorang" },
	{ id: "ljw_br:boomerang_peppermint" },
	{ id: "ljw_br:boomerang_snow" },
	{ id: "ljw_br:boomerang_applerang" },
	{ id: "ljw_br:boomerang_recoverang" }
]

function identify_rang(rang) {
	return boomerangList[rang.getProperty("ljw_br:boomerang_type")].id ?? "null";
}

system.afterEvents.scriptEventReceive.subscribe(e => {
    const entity = e.sourceEntity;
    if (e.id == "ljw_br:password") {
		const code = e.message.replaceAll(" ", "").toUpperCase();
		switch (code) {
			case "FILLYOURINVENTORYWITHRANGSFORFREE":
				entity.sendMessage("Code accepted.");
				for (let i = 1; i < boomerangList.length; i += 1)
					entity.dimension.spawnItem(new ItemStack(boomerangList[i].id), entity.location);
				return;
			default:
				entity.sendMessage("Invalid code.");
		}
	}
});

world.afterEvents.entitySpawn.subscribe(e => {
	const entity = e.entity;
	let boomerang_instance;
	
	try {
		if (entity.matches({families: [ "ljw_boomerang" ]})) {
			switch (identify_rang(entity)) {
				case "ljw_br:info_book":
					boomerang_instance = new BoomerangGuide(entity);
					break;
				case "ljw_br:boomerang_wooden":
					boomerang_instance = new BoomerangWooden(entity);
					break;
				case "ljw_br:boomerang_stone":
					boomerang_instance = new BoomerangStone(entity);
					break;
				case "ljw_br:boomerang_iron":
					boomerang_instance = new BoomerangIron(entity);
					break;
				case "ljw_br:boomerang_golden":
					boomerang_instance = new BoomerangGolden(entity);
					break;
				case "ljw_br:boomerang_diamond":
					boomerang_instance = new BoomerangDiamond(entity);
					break;
				case "ljw_br:boomerang_netherite":
					boomerang_instance = new BoomerangNetherite(entity);
					break;
				case "ljw_br:boomerang_warped":
					boomerang_instance = new BoomerangWarped(entity);
					break;
				case "ljw_br:boomerang_crimson":
					boomerang_instance = new BoomerangCrimson(entity);
					break;
				case "ljw_br:boomerang_enderang":
					boomerang_instance = new BoomerangEnderang(entity);
					break;
				case "ljw_br:boomerang_tnt":
					boomerang_instance = new BoomerangDemorang(entity);
					break;
				case "ljw_br:boomerang_frost":
					boomerang_instance = new BoomerangFrost(entity);
					break;
				case "ljw_br:boomerang_magma":
					boomerang_instance = new BoomerangMagma(entity);
					break;
				case "ljw_br:boomerang_magma_charged":
					boomerang_instance = new BoomerangMagmaCharged(entity);
					break;
				case "ljw_br:boomerang_wolfwind":
					boomerang_instance = new BoomerangWolfwind(entity);
					break;
				case "ljw_br:boomerang_gatherang":
					boomerang_instance = new BoomerangGatherang(entity);
					break;
				case "ljw_br:boomerang_hydrorang":
					boomerang_instance = new BoomerangHydrorang(entity);
					break;
				case "ljw_br:boomerang_cutterang":
					boomerang_instance = new BoomerangCutterang(entity);
					break;
				case "ljw_br:boomerang_minerang":
					boomerang_instance = new BoomerangMinerang(entity);
					break;
				case "ljw_br:boomerang_echorang":
					boomerang_instance = new BoomerangEcho(entity);
					break;
				case "ljw_br:boomerang_slime":
					boomerang_instance = new BoomerangSlime(entity);
					break;
				case "ljw_br:boomerang_gilded":
					boomerang_instance = new BoomerangGilded(entity);
					break;
				case "ljw_br:boomerang_turborang":
					boomerang_instance = new BoomerangTurborang(entity);
					break;
				case "ljw_br:boomerang_void":
					boomerang_instance = new BoomerangHolorang(entity);
					break;
				case "ljw_br:boomerang_enigmarang":
					const choice = Math.floor(Math.random() * 10);
					if (choice == 0)
						boomerang_instance = new BoomerangNetherite(entity);
					else if (choice == 1)
						boomerang_instance = new BoomerangDemorang(entity);
					else if (choice == 2)
						boomerang_instance = new BoomerangSnow(entity);
					else if (choice == 3)
						boomerang_instance = new BoomerangTurborang(entity);
					else if (choice == 4)
						boomerang_instance = new BoomerangSlime(entity);
					else if (choice == 5)
						boomerang_instance = new BoomerangWolfwind(entity);
					else if (choice == 6)
						boomerang_instance = new BoomerangWarped(entity);
					else
						boomerang_instance = new BoomerangIron(entity);
					break;
				case "ljw_br:boomerang_breezyrang":
					boomerang_instance = new BoomerangBreezyrang(entity);
					break;
				case "ljw_br:boomerang_raptorang":
					boomerang_instance = new BoomerangRaptorang(entity);
					break;
				case "ljw_br:boomerang_heavenlyrang":
					boomerang_instance = new BoomerangHeavenlyrang(entity);
					break;
				case "ljw_br:boomerang_jellyrang":
					boomerang_instance = new BoomerangJellyrang(entity);
					break;
				case "ljw_br:boomerang_torcharang":
					boomerang_instance = new BoomerangWooden(entity);
					break;
				case "ljw_br:boomerang_hopperang":
					boomerang_instance = new BoomerangHopperang(entity);
					break;
				case "ljw_br:boomerang_collectorang":
					boomerang_instance = new BoomerangCollectorang(entity);
					break;
				case "ljw_br:boomerang_peppermint":
					boomerang_instance = new BoomerangPeppermint(entity);
					break;
				case "ljw_br:boomerang_snow":
					boomerang_instance = new BoomerangSnow(entity);
					break;
				case "ljw_br:boomerang_applerang":
					boomerang_instance = new BoomerangApplerang(entity);
					break;
				case "ljw_br:boomerang_recoverang":
					boomerang_instance = new BoomerangRecoverang(entity);
					break;
			}
		}
	}
	catch(err) {
		debug.error(`Boomerang init error: ${err}`);
	}
});

world.afterEvents.itemUse.subscribe(e => {
	const target = e.source;
	const itemStack = e.itemStack;
	if (target == null)
		return;
	
	try {
		//if (itemStack.hasTag("ljw_br:boomerang")) {
		//	target.dimension.spawnEntity(`ljw_br:boomerang<${itemStack.typeId}>`, {x: target.location.x, y: target.location.y + 1, z: target.location.z});
		//}
		
		if (itemStack.typeId == "ljw_br:boomerang_void") {
			target.dimension.spawnEntity("ljw_br:boomerang<ljw_br:boomerang_void>", {x: target.location.x, y: target.location.y + 1, z: target.location.z});
		}
		else if (itemStack.typeId == "ljw_br:boomerang_hydrorang") {
			if (!target.isInWater)
				target.sendMessage("This boomerang can only be used underwater!");
			else {
				target.dimension.spawnEntity("ljw_br:boomerang<ljw_br:boomerang_hydrorang>", {x: target.location.x, y: target.location.y + 1.5, z: target.location.z});
				target.getComponent("equippable").setEquipment("Mainhand");
			}
		}
		else if (itemStack.typeId == "ljw_br:info_book") {
			if (target.isSneaking) {
				target.dimension.spawnEntity("ljw_br:boomerang<ljw_br:boomerang_guide>", {x: target.location.x, y: target.location.y + 1, z: target.location.z});
				util.decrement_item(target, "ljw_br:info_book", 1);
			}
		}
	}
	catch(err) {
		debug.error(`Boomerang use error [${itemStack.typeId}]: ${err}`)
	}
});

world.afterEvents.projectileHitEntity.subscribe(e => {
	try {
		if (e.projectile?.typeId == "ljw_br:boomerang") {
			const rang = e.projectile;
			const target = e.getEntityHit().entity;
			try {
				if(target.id != rang.getDynamicProperty("ljw_br:owner_id")) {
					target.applyDamage(rang_roster[rang.id].damage, {damagingEntity: rang_roster[rang.id].owner, damagingProjectile: rang});
					if (target.typeId != "ljw_br:boomerang_target")
						target.applyKnockback(rang.getVelocity().x, rang.getVelocity().z, rang_roster[rang.id].knockback, rang_roster[rang.id].vertical_knockback);
					debug.info("Boomerang Hit Entity!");
					rang.setDynamicProperty(`ljw_br:hitObject`, true);
				}
			}
			catch(err) {
				debug.warn(`Rang damage error: ${err}`);
			}
			
			// Boomerang variant effects
			switch (identify_rang(rang)) {
				case "ljw_br:boomerang_frost":
					try {
						if(target.id != rang.getDynamicProperty("ljw_br:owner_id"))
							target.addEffect("slowness", 200, {amplifier: 2});
					}
					catch(err) {
						debug.warn(`Frost Boomerang Error: ${err}`);
					}
					break;
				case "ljw_br:boomerang_magma_charged":
					try {
						if(target.id != rang.getDynamicProperty("ljw_br:owner_id"))
							target.setOnFire(5);
					}
					catch(err) {
						debug.warn(`Magmarang Error: ${err}`);
					}
					break;
				case "ljw_br:boomerang_heavenlyrang":
					try {
						if(target.id != rang.getDynamicProperty("ljw_br:owner_id")) {
							const choice = Math.floor(Math.random() * 5);
							if (choice == 0)
								target.addEffect("speed", 200, {amplifier: 2});
							else if (choice == 1)
								target.addEffect("jump_boost", 200, {amplifier: 2});
							else if (choice == 2)
								target.addEffect("regeneration", 200, {amplifier: 2});
							else if (choice == 3)
								target.addEffect("resistance", 200, {amplifier: 2});
							else if (choice == 4)
								target.addEffect("strength", 200, {amplifier: 2});
						}
					}
					catch(err) {
						debug.warn(`Heavenlyrang Error: ${err}`);
					}
					break;
				case "ljw_br:boomerang_peppermint":
					try {
						if(target.id != rang.getDynamicProperty("ljw_br:owner_id"))
							rang_roster[rang.id].owner.addEffect("speed", 100, {amplifier: 2});
					}
					catch(err) {
						debug.warn(`Peppermint Boomerang Error: ${err}`);
					}
					break;
				case "ljw_br:boomerang_applerang":
					try {
						if(target.id != rang.getDynamicProperty("ljw_br:owner_id"))
							target.applyKnockback(0, 0, 0, -10);
					}
					catch(err) {
						debug.warn(`Applerang Error: ${err}`);
					}
					break;
				case "ljw_br:boomerang_recoverang":
					try {
						if(target.id != rang.getDynamicProperty("ljw_br:owner_id")) {
							const healthComponent = target.getComponent("health");
							if (healthComponent != null)
								healthComponent.setCurrentValue(Math.min(healthComponent.currentValue + 5, healthComponent.effectiveMax));
						}
					}
					catch(err) {
						debug.warn(`Recoverang Error: ${err}`);
					}
					break;
			}
		}
	}
	catch (err)
	{}
});

world.afterEvents.projectileHitBlock.subscribe(e => {
	const projectile = e.projectile;
	try {
		if (projectile.matches({families: [ "ljw_boomerang" ]}))
		{
			try {
				const rang = projectile;
				rang.setDynamicProperty(`ljw_br:hitObject`, true);
				rang_roster[rang.id].hit_block(e.getBlockHit().block);
			}
			catch(err)
			{}
		}
	}
	catch (err)
	{}
});

let rang_roster_init = false;
let rang_roster = {};

class Boomerang {
	rang;
	id;
	owner;
	lifetime = 10;
	step = 0;
	speed = 1.5;
	damage = 2;
	knockback = 0.5;
	vertical_knockback = 0;
	return_speed = this.speed;
	x_angle;
	y_angle;
	z_angle;
	x_change;
	y_change;
	z_change;
	return_steps = 27;
	hit = false;
	hit_count = 0;
	hit_step = 0;
	stop = false;
	returning = false;
	rang_given = false;
	travels_air = true;
	travels_water = false;
	invalid_area_tolerance = 5;
	return_speed_modifier = 10;
	tick_instance;
	
	constructor(item) {
		this.rang = item;
		this.id = item.id;
		this.owner = this.rang.dimension.getPlayers({ location: { x: item.location.x, y: item.location.y, z: item.location.z }, closest: 1 })[0];
		this.x_angle = this.owner.getViewDirection().x;
		this.y_angle = this.owner.getViewDirection().y;
		this.z_angle = this.owner.getViewDirection().z;
		
		if ((this.x_angle >= 0) && (this.z_angle >= 0))
		{
			this.x_change = 1;
			this.z_change = -1;
		}
		else if ((this.x_angle >= 0) && (this.z_angle < 0))
		{
			this.x_change = -1;
			this.z_change = -1;
		}
		else if ((this.x_angle < 0) && (this.z_angle < 0))
		{
			this.x_change = -1;
			this.z_change = 1;
		}
		else if ((this.x_angle < 0) && (this.z_angle >= 0))
		{
			this.x_change = 1;
			this.z_change = 1;
		}
		
		this.y_change = (-2 * this.y_angle) / this.return_steps;
		
		this.rang.setDynamicProperty(`ljw_br:owner_id`, this.owner.id);
		rang_roster[item.id] = (this);
		this.tick_instance = system.runInterval(() => this.tick());
		debug.info(`New ${identify_rang(this.rang)} [${this.rang.getProperty("ljw_br:boomerang_type")}] from "${this.owner.name}"`);
		//debug.info(`Hello World! My name is ${this.id} and the player that threw me is "${this.owner.name}"`);
		//debug.info(`Owner rotation is ${this.owner.getViewDirection().x} ${this.owner.getViewDirection().y} ${this.owner.getViewDirection().z}`);
	}
	tick() {
		try {
			this.go();
		}
		catch(err) {
			debug.error(`Boomerang error: ${err}`);
			this.unload();
		}
	}
	go() {
		this.hit = false;
		this.step_forward();
		
		this.environment_check();
		this.hit_detect();
		this.flight_sound();
		
		if ((this.step >= this.lifetime) && (this.step < this.lifetime + this.return_steps))
		{
			this.curve();
		}
		else if (this.step >= this.lifetime + this.return_steps)
		{
			this.returning = true;
			this.return_step();
		}
	}
	unload() {
		system.clearRun(this.tick_instance);
	}
	flight_sound() {
		if (this.step % 2 == 0)
			world.playSound("random.bow", this.rang.location, {pitch: 0.9 + 0.2 * this.speed});
	}
	environment_check() {
		//this.rang.dimension.getBlockFromRay({ x: this.rang.location.x, y: this.rang.location.y, z: this.rang.location.z }, { x: (this.speed * this.x_angle), y: (this.speed * this.y_angle), z: (this.speed * this.z_angle) });
		if (!this.travels_air && !this.rang.isInWater)
		{
			if (this.invalid_area_tolerance <= 0)
				this.fail();
			else
				this.invalid_area_tolerance -= 1;
		}
		if (!this.travels_water && this.rang.isInWater)
		{
			if (this.invalid_area_tolerance <= 0)
				this.rang.setDynamicProperty("ljw_br:hitObject", true);
			else
				this.invalid_area_tolerance -= 1;
		}
	}
	hit_detect() {
		//this.rang.dimension.getBlockFromRay({ x: this.rang.location.x, y: this.rang.location.y, z: this.rang.location.z }, { x: (this.speed * this.x_angle), y: (this.speed * this.y_angle), z: (this.speed * this.z_angle) });
		
		if (this.rang.getDynamicProperty("ljw_br:hitObject") == true)
		{
			this.hit_count += 1;
			this.hit = true;
			this.rang.setDynamicProperty(`ljw_br:hitObject`, false);
			
			if (this.hit_step == 0)
				this.hit_step = this.step;
		}
		
		if (this.hit_step > 0)
			this.hit_event();
	}
	hit_block(block) { }
	hit_event() {
		this.x_angle = 0;
		this.y_angle = 2;
		this.z_angle = 0;
		this.stop = true;
		this.returning = true;
		
		if (this.step > (this.hit_step + 5))
			this.returned();
	}
	step_forward() {
		this.rang.clearVelocity();
		
		if (this.step < this.lifetime + this.return_steps)
		{
			this.rang.applyImpulse({ x: (this.speed * this.x_angle), y: (this.speed * this.y_angle), z: (this.speed * this.z_angle) });
		}
		else
		{
			const intensity = Math.min(10, util.distance3D(this.rang.location, this.owner.location)) / this.return_speed_modifier;
			this.rang.applyImpulse({ x: (this.return_speed * this.x_angle * intensity), y: (this.return_speed * this.y_angle * intensity), z: (this.return_speed * this.z_angle * intensity) });
		}
		
		this.step += 1;
	}
	curve() {
		if (!this.stop)
		{
			this.x_angle = this.x_angle + 0.1 * this.x_change;
			this.y_angle += this.y_change;
			this.z_angle = this.z_angle + 0.1 * this.z_change;
			
			// Adjust change when bounds are hit
			if (this.x_angle >= 1.0)
				this.x_change = -1;
			else if (this.x_angle <= -1.0)
				this.x_change = 1;
			if (this.z_angle >= 1.0)
				this.z_change = -1;
			else if (this.z_angle <= -1.0)
				this.z_change = 1;
			
			// Clamp rotation values
			if (this.x_angle > 1.0)
				this.x_angle = 1.0 - (this.x_angle - 1);
			else if (this.x_angle < -1.0)
				this.x_angle = -1.0 + (this.x_angle + 1);
			if (this.z_angle > 1.0)
				this.z_angle = 1.0 - (this.z_angle - 1);
			else if (this.z_angle < -1.0)
				this.z_angle = -1.0 + (this.z_angle + 1);
		}
		
		//debug.info(`rotation is ${this.x_angle}`)
	}
	return_step() {
		this.x_angle = util.clamp((this.owner.location.x - this.rang.location.x), -0.4, 0.4);
		
		if (Math.floor(this.owner.location.y + 0.8) != Math.floor(this.rang.location.y))
			this.y_angle = util.clamp((this.owner.location.y + 0.8 - this.rang.location.y), -0.4, 0.4);
		else
			this.y_angle = 0;
		
		this.z_angle = util.clamp((this.owner.location.z - this.rang.location.z), -0.4, 0.4);
		
		const distance_to_owner = util.distance3D(this.rang.location, this.owner.location);
		
		if (this.step % 10 == 0)
			this.rang.teleport({x: this.rang.location.x + this.x_angle, y: this.rang.location.y + this.y_angle, z: this.rang.location.z + this.z_angle})
		
		if(distance_to_owner < 1.5)
			this.returned();
	}
	returned() {
		debug.info(`${identify_rang(this.rang)} returned`);
		this.rang.triggerEvent(`ljw_br:event_despawn`);
		this.give_rang();
		this.unload();
	}
	fail() {
		try
		{
			this.owner = world.getEntity(this.rang.getDynamicProperty("ljw_br:owner_id"));
			this.give_rang();
		}
		catch(err)
		{
			debug.error(`Boomerang epic fail: ${err}`)
			this.epic_fail();
		}
		this.rang.triggerEvent(`ljw_br:event_despawn`);
		this.unload();
	}
	epic_fail() {
		this.rang.dimension.spawnItem(new ItemStack(identify_rang(this.rang)), this.rang.location);
	}
	give_rang() {
		if (!this.rang_given)
		{
			const container = this.owner.getComponent("inventory").container;
			if (container.emptySlotsCount > 0)
				container.addItem(new ItemStack(identify_rang(this.rang)));
			else
				this.epic_fail();
		}
		
		this.rang_given = true;
	}
}

class BoomerangGuide extends Boomerang {
	speed = 0.65;
	damage = 1;
	knockback = 0;
	epic_fail() {
		this.rang.dimension.spawnItem(new ItemStack("ljw_br:info_book"), this.rang.location);
	}
	give_rang() {
		if (!this.rang_given)
		{
			const container = this.owner.getComponent("inventory").container;
			if (container.emptySlotsCount > 0)
				container.addItem(new ItemStack("ljw_br:info_book"));
			else
				this.epic_fail();
		}
		
		this.rang_given = true;
	}
}

class BoomerangWooden extends Boomerang {
	speed = 0.65;
	damage = 2;
	knockback = 0.25;
}

class BoomerangStone extends Boomerang {
	speed = 0.8;
	damage = 3;
	knockback = 0.5;
}

class BoomerangIron extends Boomerang {
	speed = 1.25;
	damage = 5;
	knockback = 0.65;
}

class BoomerangGolden extends Boomerang {
	speed = 1.5;
	damage = 4;
	knockback = 0.1;
	lifetime = 2;
}

class BoomerangDiamond extends Boomerang {
	speed = 1.5;
	damage = 6;
	knockback = 1.0;
}

class BoomerangNetherite extends Boomerang {
	speed = 1.75;
	damage = 6.5;
	knockback = 1.25;
	vertical_knockback = 0.4;
	travels_water = true;
}

class BoomerangWarped extends Boomerang {
	speed = 0.1;
	return_speed = 1.0;
	damage = 2;
	lifetime = 120;
	return_steps = 1;
	hit_event() {
	}
}

class BoomerangCrimson extends Boomerang {
	speed = 0.65;
	damage = 2;
	knockback = 0.25;
	hit_event() {
	}
}

class BoomerangHydrorang extends Boomerang {
	speed = 0.65;
	damage = 5;
	knockback = 0.25;
	travels_water = true;
	travels_air = false;
	return_speed_modifier = 7;
}

class BoomerangEnderang extends Boomerang {
	speed = 0.8;
	damage = 3;
	lifetime = 35;
	travels_water = true;
	warp() {
		world.playSound("mob.shulker.teleport", this.owner.location);
		this.owner.teleport(this.rang.location);
		system.runTimeout(() => {world.playSound("mob.shulker.teleport", this.owner.location);}, 2);
		this.returned();
	}
	curve() {
		this.warp();
	}
	hit_event() {
		this.warp();
	}
}

class BoomerangDemorang extends Boomerang {
	speed = 0.5;
	lifetime = 35;
	explode() {
		//this.rang.dimension.createExplosion({ x: this.rang.location.x, y: this.rang.location.y, z: this.rang.location.z }, 3);
		this.rang.triggerEvent(`ljw_br:event_explode`);
		this.returned();
	}
	curve() {
		this.explode();
	}
	hit_event() {
		this.explode();
	}
}

class BoomerangFrost extends Boomerang {
	speed = 0.75;
	lifetime = 15;
	damage = 3;
	step_forward() {
		super.step_forward();
		const block = this.rang.dimension.getBlock(this.rang.location);
		if ((this.rang.isInWater == true) && (this.hit_count == 0))
		{
			this.speed = 0;
			this.rang.clearVelocity();
			const blockLocation = {x: Math.floor(this.rang.location.x), y: Math.floor(this.rang.location.y), z: Math.floor(this.rang.location.z)};
			const y = 0;
			for (let x = -1; x <= 1; x += 1)
				for (let z = -1; z <= 1; z += 1)
				{
					let testLocation = {x: blockLocation.x + x, y: blockLocation.y + y, z: blockLocation.z + z};
					if ((this.rang.dimension.getBlock(testLocation)?.permutation == BlockPermutation.resolve("minecraft:water")) || (this.rang.dimension.getBlock(testLocation)?.permutation == BlockPermutation.resolve("minecraft:flowing_water")))
					{
						this.rang.dimension.getBlock(testLocation).setPermutation(BlockPermutation.resolve("minecraft:frosted_ice"));
						this.rang.dimension.spawnParticle("ljw_br:frost_blast", this.rang.location);
					}
				}
			this.rang.setDynamicProperty("ljw_br:hitObject", true);
		}
		else if (block.permutation == BlockPermutation.resolve("minecraft:fire"))
			block.setPermutation(BlockPermutation.resolve("minecraft:air"))
		else if (block.permutation == BlockPermutation.resolve("minecraft:lava"))
		{
			block.setPermutation(BlockPermutation.resolve("minecraft:obsidian"))
			this.speed = 0;
			this.rang.clearVelocity();
			this.rang.setDynamicProperty("ljw_br:hitObject", true);
		}
	}
}

class BoomerangGilded extends Boomerang {
	speed = 0.8;
	damage = 5;
	lifetime = 7;
	return_steps = 70;
}

class BoomerangHolorang extends Boomerang {
	speed = 0.75;
	damage = 3.0;
	give_rang() {
	}
	hit_event() {
	}
}

class BoomerangMagma extends Boomerang {
	speed = 0.75;
	damage = 4;
	state = "cold";
	hit_block(block) {
		if (block.permutation == BlockPermutation.resolve("minecraft:magma"))
			this.state = "hot";
	}
	step_forward() {
		super.step_forward();
		if (this.rang.isInWater == true)
			this.state = "cold";
		else if (this.rang.hasComponent("minecraft:is_ignited"))
			this.state = "hot";
	}
	give_rang() {
		if (!this.rang_given)
		{
			const container = this.owner.getComponent("inventory").container;
			if (container.emptySlotsCount > 0)
			{
				if (this.state == "hot")
					this.owner.getComponent("inventory").container.addItem(new ItemStack(`ljw_br:boomerang_magma_charged`));
				else
					this.owner.getComponent("inventory").container.addItem(new ItemStack(`ljw_br:boomerang_magma`));
			}
			else
			{
				if (this.state == "hot")
					this.rang.dimension.spawnItem(new ItemStack(`ljw_br:boomerang_magma_charged`), this.rang.location);
				else
					this.rang.dimension.spawnItem(new ItemStack(`ljw_br:boomerang_magma`), this.rang.location);
			}
		}
		
		this.rang_given = true;
	}
}

class BoomerangMagmaCharged extends Boomerang {
	speed = 1.25;
	damage = 5.5;
	state = "hot";
	hit_block(block) {
		if (block.permutation == BlockPermutation.resolve("minecraft:magma"))
			this.state = "hot";
		else if (block.permutation == BlockPermutation.resolve("minecraft:ice"))
			this.block.setPermutation(BlockPermutation.resolve("minecraft:water"))
		else if (block.permutation == BlockPermutation.resolve("minecraft:frosted_ice"))
			this.block.setPermutation(BlockPermutation.resolve("minecraft:water"))
	}
	step_forward() {
		super.step_forward();
		if (this.rang.isInWater == true)
			this.state = "cold";
		else if (this.rang.hasComponent("minecraft:is_ignited"))
			this.state = "hot";
	}
	give_rang() {
		if (!this.rang_given)
		{
			const container = this.owner.getComponent("inventory").container;
			if (container.emptySlotsCount > 0)
			{
				if (this.state == "hot")
					this.owner.getComponent("inventory").container.addItem(new ItemStack(`ljw_br:boomerang_magma_charged`));
				else if (this.state == "cold")
					this.owner.getComponent("inventory").container.addItem(new ItemStack(`ljw_br:boomerang_magma`));
				else
				{
					if (Math.random() < 0.97)
						this.owner.getComponent("inventory").container.addItem(new ItemStack(`ljw_br:boomerang_magma_charged`));
					else
						this.owner.getComponent("inventory").container.addItem(new ItemStack(`ljw_br:boomerang_magma`));
				}
			}
			else
			{
				if (this.state == "hot")
					this.rang.dimension.spawnItem(new ItemStack(`ljw_br:boomerang_magma_charged`), this.rang.location);
				else if (this.state == "cold")
					this.rang.dimension.spawnItem(new ItemStack(`ljw_br:boomerang_magma`), this.rang.location);
				else
				{
					if (Math.random() < 0.97)
						this.rang.dimension.spawnItem(new ItemStack(`ljw_br:boomerang_magma_charged`), this.rang.location);
					else
						this.rang.dimension.spawnItem(new ItemStack(`ljw_br:boomerang_magma`), this.rang.location);
				}
			}
		}
		
		this.rang_given = true;
	}
}

class BoomerangSlime extends Boomerang {
	speed = 0.65;
	lifetime = 10;
	damage = 0;
	hit_event() {
		if (this.hit == true)
		{
			const mob_list = this.rang.dimension.getEntities({ location: { x: this.rang.location.x, y: this.rang.location.y, z: this.rang.location.z }, closest: 2, maxDistance: 2, excludeFamilies: ["player", "ljw_boomerang", "boomerang_target"] });
			if (mob_list.length > 0)
			{
				if (this.step < this.lifetime + this.return_steps)
				{
					mob_list[0].applyImpulse({ x: (this.speed * this.x_angle * 5), y: (this.speed * this.y_angle * 5), z: (this.speed * this.z_angle * 5) });
				}
				else
				{
					mob_list[0].applyImpulse({ x: (this.return_speed * this.x_angle * 5), y: (this.return_speed * this.y_angle * 5), z: (this.return_speed * this.z_angle * 5) });
				}
			}
		}
	}
}

class BoomerangCutterang extends Boomerang {
	speed = 0.65;
	damage = 3;
	hit_event() {
	}
}

class BoomerangMinerang extends Boomerang {
	speed = 0.65;
	damage = 3;
	hit_event() {
	}
}

class BoomerangTurborang extends Boomerang {
	speed = 2.25;
	damage = 0;
	lifetime = 25;
	
	go() {
		this.damage = 20 * (Math.E ** (util.clamp(this.step/this.lifetime, 0, 1) - 1));
		super.go();
	}
}

class BoomerangEcho extends Boomerang {
	speed = 0.17;
	lifetime = 1200;
	damage = 7;
	return_steps = 1;
	cancel = false;
	
	constructor(item) {
		super(item);
		this.owner.addTag(`echorang_${this.rang.id}`);
		this.owner.runCommandAsync(`titleraw @s actionbar {"rawtext":[{"text":"Use the camera to aim!\n Press Sneak to exit."}]}`);
	}
	go() {
		this.hit = false;
		this.step_forward();
		
		this.hit_detect();
		
		const ownerDistance = util.distance3D(this.rang.location, this.owner.location);
		const ownerDistanceLimit = 700;
		const limitMarkerFilled = "|";
		const limitMarkerEmpty = ".";
		const rangeBar = `${limitMarkerFilled.repeat(util.clamp(Math.ceil(10 * ownerDistance/ownerDistanceLimit),0,10))}${limitMarkerEmpty.repeat(util.clamp(10 - Math.ceil(10 * ownerDistance/ownerDistanceLimit),0,10))}`;
		
		if ((ownerDistance > ownerDistanceLimit) || this.owner.isSneaking || this.owner.isFalling || ((this.lifetime - this.step) <= 0))
			this.rang.setDynamicProperty(`ljw_br:hitObject`, true);
		
		if (this.step > 60)
			this.owner.runCommandAsync(`titleraw @s actionbar {"rawtext":[{"text":"Range Limit: ${rangeBar}\nTime Limit: ${Math.max((this.lifetime/20) - (this.step/20), 0).toFixed(2)}s"}]}`);
		
		if (!this.cancel)
		{
			this.curve();
		}
		else
		{
			this.returning = true;
			this.return_step();
		}
	}
	step_forward() {
		this.rang.clearVelocity();
		const viewDir = this.owner.getViewDirection();
		const baseZoom = 2.5;
		
		if (this.returning == false)
			this.rang.runCommandAsync(`camera "${this.owner.name}" set minecraft:free ease 0.5 linear pos ${this.rang.location.x - baseZoom * Math.sin(Math.PI * viewDir.x / 4)} ${1 + this.rang.location.y - 2.5 * viewDir.y} ${this.rang.location.z - baseZoom * Math.sin(Math.PI * viewDir.z / 4)} facing ${this.rang.location.x + viewDir.x * 2} ${this.rang.location.y} ${this.rang.location.z + viewDir.z * 2}`);
		else
			this.rang.runCommandAsync(`camera "${this.owner.name}" clear`);
			
		if (this.step < this.lifetime + this.return_steps)
		{
			this.rang.applyImpulse({ x: (this.speed * this.x_angle), y: (this.speed * this.y_angle), z: (this.speed * this.z_angle) });
		}
		else
		{
			this.rang.applyImpulse({ x: (this.return_speed * this.x_angle), y: (this.return_speed * this.y_angle), z: (this.return_speed * this.z_angle) });
		}
		
		this.step += 1;
	}
	curve() {
		if (!this.stop)
		{
			this.x_angle = this.owner.getViewDirection().x;
			this.y_angle = this.owner.getViewDirection().y;
			this.z_angle = this.owner.getViewDirection().z;
		}
		
		//debug.info(`rotation is ${this.x_angle}`);
	}
}

class BoomerangWolfwind extends Boomerang {
	target_type = "monster";
	target_list = [];
	nearby_list = [];
	target_num = 0;
	bounce = 0;
	speed = 0.8;
	damage = 5;
	return_speed = 0.8;
	travels_water = true;
	hit_bounce = 0.75;
	return_speed_modifier = 7;
	
	constructor(item) {
		super(item);
		this.target_list = [...this.rang.dimension.getEntities({ families: [ this.target_type ], location: { x: item.location.x, y: item.location.y, z: item.location.z }, maxDistance: 25 }), ...this.rang.dimension.getEntities({ families: [ "boomerang_target" ], location: { x: item.location.x, y: item.location.y, z: item.location.z }, maxDistance: 25 })];
	}
	go() {
		this.hit = false;
		this.step_forward();
		
		this.hit_detect();
		
		if (this.target_num < this.target_list.length)
		{
			this.curve();
		}
		else
		{
			this.returning = true;
			this.return_step();
		}
	}
	curve() {
		try
		{
			let distance = util.clamp(Math.pow(Math.pow(this.target_list[this.target_num].location.x - this.rang.location.x, 2) + Math.pow(this.target_list[this.target_num].location.z - this.rang.location.z, 2),0.5),0,15);
			
			this.x_angle = util.clamp((this.target_list[this.target_num].location.x - this.rang.location.x), -0.4, 0.4);
			
			if (Math.floor(this.target_list[this.target_num].location.y + 0.5) != Math.floor(this.rang.location.y))
				this.y_angle = util.clamp((this.target_list[this.target_num].location.y + 0.5 - this.rang.location.y), -0.4, 0.4);
			else
				this.y_angle = 0;
			
			if (this.rang.location.y < this.target_list[this.target_num].location.y + 7)
				this.y_angle = util.clamp(this.y_angle + (0.05 * distance), -0.4, 0.4);
			
			this.z_angle = util.clamp((this.target_list[this.target_num].location.z - this.rang.location.z), -0.4, 0.4);
			
			this.nearby_list = this.rang.dimension.getEntities({ excludeFamilies: ["player", "ljw_boomerang"], location: { x: this.rang.location.x, y: this.rang.location.y, z: this.rang.location.z }, maxDistance: 1.5 });
			
			if (this.bounce > 0)
			{
				this.bounce = Math.max(this.bounce - 0.1, 0);
			}
			
			if(this.nearby_list.length > 0)
			{
				if(this.nearby_list.indexOf(this.target_list[this.target_num]) > -1)
				{
					this.target_num += 1;
				}
			}
		}
		catch(err)
		{
			this.target_num += 1;
		}
	}
	return_step() {
		let distance = Math.pow(Math.pow(this.owner.location.x - this.rang.location.x, 2) + Math.pow(this.owner.location.z - this.rang.location.z, 2),0.5);
		
		this.x_angle = util.clamp((this.owner.location.x - this.rang.location.x), -0.4, 0.4);
		
		if (Math.floor(this.owner.location.y + 0.8) != Math.floor(this.rang.location.y))
			this.y_angle = util.clamp((this.owner.location.y + 0.8 - this.rang.location.y), -0.4, 0.4);
		else
			this.y_angle = 0;
		
		if (this.rang.location.y < this.owner.location.y + 7)
			this.y_angle = util.clamp(this.y_angle + (0.05 * distance), -0.4, 0.4);
		
		this.z_angle = util.clamp((this.owner.location.z - this.rang.location.z), -0.4, 0.4);
		
		const distance_to_owner = Math.pow(Math.pow(this.owner.location.x - this.rang.location.x, 2) + Math.pow(this.owner.location.y - this.rang.location.y, 2) + Math.pow(this.owner.location.z - this.rang.location.z, 2),0.5);
		
		if(distance_to_owner < 1.5)
			this.returned();
	}
	hit_event() {
		if (this.hit)
			this.bounce = this.hit_bounce;
	}
}

class BoomerangCollectorang extends BoomerangWolfwind {
	target_type = "item";
	damage = 0;
	hit_bounce = 0.0;
	constructor(item) {
		item.runCommand(`tag @e[type=minecraft:item,r=20] add item`);
		super(item);
		this.target_list = this.rang.dimension.getEntities({ tags: [ this.target_type ], location: { x: item.location.x, y: item.location.y, z: item.location.z }, maxDistance: 25 });
	}
	go() {
		this.rang.runCommandAsync(`tp @e[type=minecraft:item,r=2] ~ ~ ~`);
		this.rang.runCommandAsync(`tag @e[type=minecraft:item,r=1] remove item`);
		super.go();
	}
	returned() {
		this.rang.runCommand(`tp @e[type=minecraft:item,r=2] "${this.owner.nameTag}"`);
		super.returned();
	}
}

class BoomerangHopperang extends BoomerangIron {
	damage = 0;
	speed = 0.8;
	hit_bounce = 0.0;
	constructor(item) {
		super(item);
	}
	go() {
		this.rang.runCommandAsync(`tp @e[type=minecraft:item,r=2] ~ ~ ~`);
		this.rang.runCommandAsync(`tag @e[type=minecraft:item,r=1] remove item`);
		super.go();
	}
	hit_event() {}
	returned() {
		this.rang.runCommand(`tp @e[type=minecraft:item,r=2] "${this.owner.nameTag}"`);
		super.returned();
	}
}

class BoomerangGatherang extends BoomerangWolfwind {
	target_type = "item";
	damage = 0;
	hit_bounce = 0.0;
	constructor(item) {
		super(item);
		const itemList = [];
		const container = this.owner.getComponent("inventory").container;
		for (let i = 0; i < container.size; i += 1) {
			const itemStack = container.getItem(i);
			if (itemStack != null)
				itemList.push(itemStack.typeId);
		}
		for (const item of this.owner.dimension.getEntities({type: "minecraft:item", location: this.owner.location, maxDistance: 20})) {
			if (itemList.indexOf(`${item.getComponent("minecraft:item")?.itemStack.typeId}`) !=  -1)
				item.runCommand(`tag @s add item`);
		}
		this.target_list = this.rang.dimension.getEntities({ tags: [ this.target_type ], location: { x: item.location.x, y: item.location.y, z: item.location.z }, maxDistance: 25 });
	}
	go() {
		this.rang.runCommandAsync(`tp @e[type=minecraft:item,r=2] ~ ~ ~`);
		this.rang.runCommandAsync(`tag @e[type=minecraft:item,r=1] remove item`);
		super.go();
	}
	returned() {
		this.rang.runCommand(`tp @e[type=minecraft:item,r=2] "${this.owner.nameTag}"`);
		super.returned();
	}
}

class BoomerangBreezyrang extends Boomerang {
	speed = 0.8;
	damage = 3;
	knockback = 0.5;
	blastRange = 3.5;
	horizontalIntensity = 3.5;
	verticalIntensity = 1.35;
	hitMem = 0;
	
	go() {
		const particleLocVariance = 0.5;
		this.rang.dimension.spawnParticle("minecraft:wind_charged_emitter", {x: this.rang.location.x - particleLocVariance + Math.random() * particleLocVariance * 2, y: this.rang.location.y - particleLocVariance + Math.random() * particleLocVariance * 2, z: this.rang.location.z - particleLocVariance + Math.random() * particleLocVariance * 2});
		super.go();
	}
	
	hit_event() {
		if ((!this.returning) && (this.hit_count > this.hitMem)) {
			this.rang.dimension.spawnParticle("minecraft:wind_explosion_emitter", this.rang.location);
			world.playSound("wind_charge.burst", this.rang.location);
			for (const entity of this.rang.dimension.getEntities({location: this.rang.location, maxDistance: this.blastRange}))
			{
				if (entity == this.rang)
					continue;
				
				try {
					const xIntensity = 1 - Math.abs((this.rang.location.x - entity.location.x) / this.blastRange);
					const zIntensity = 1 - Math.abs((this.rang.location.z - entity.location.z) / this.blastRange);
					const intensity = (xIntensity + zIntensity) / 2;
					entity.applyKnockback((entity.location.x - this.rang.location.x) / this.blastRange, (entity.location.z - this.rang.location.z) / this.blastRange, this.horizontalIntensity * intensity, this.verticalIntensity * intensity);
				}
				catch (err) {}
			}
			this.hitMem = this.hit_count;
		}
		super.hit_event();
	}
}

class BoomerangRaptorang extends Boomerang {
	speed = 0.5;
	damage = 5;
	knockback = 0.2;
	lifetime = 1;
	return_steps = 15;
	return_speed = 1.5;
	
	hit_event() {}
}

class BoomerangJellyrang extends BoomerangIron {
	damage = 5;
	travels_water = true;
}

class BoomerangPeppermint extends Boomerang {
	damage = 2;
}

class BoomerangSnow extends Boomerang {
	damage = 0.001;
}

class BoomerangApplerang extends Boomerang {
	damage = 0;
}

class BoomerangRecoverang extends Boomerang {
	damage = 0;
}

class BoomerangLightning extends Boomerang {
	speed = 0.8;
	damage = 5;
	go() {
		//if (this.rang.getProperty("ljw_br:in_rain") && (this.step % 10 == 0))
		if ((this.step % 5 == 0) && (util.distance3D(this.rang.location, this.owner.location) > 10))
			this.rang.dimension.spawnEntity("minecraft:lightning_bolt", this.rang.dimension.getBlockFromRay(this.rang.location, {x: 0, y: -1, z: 0}).block.location);
		
		super.go();
	}
	hit_event() {}
}

class BoomerangHeavenlyrang extends Boomerang {
	speed = 0.8;
	damage = 0;
}

class BoomerangRavager extends Boomerang {
	speed = 0.5;
	lifetime = 35;
	damage = 0;
	explode() {
		const distanceMax = 7;
		const horizontalIntensity = 3.5;
		const verticalIntensity = 1.5;
		const list = this.rang.dimension.getEntities({location: this.rang.location, maxDistance: distanceMax});
		this.rang.dimension.spawnParticle("minecraft:knockback_roar_particle", this.rang.location);
		world.playSound("mob.ravager.roar", this.rang.location);
		for (const entity of list)
		{
			try {
				const intensity = horizontalIntensity * (1 - (util.distance2Dz(this.rang.location, entity.location) / 22.5));
				entity.applyKnockback((entity.location.x - this.rang.location.x) / distanceMax, (entity.location.z - this.rang.location.z) / distanceMax, intensity, (this.rang.location.y > entity.location.y ? -1 * verticalIntensity : verticalIntensity) * (1 - Math.abs(entity.location.y - this.rang.location.y) / distanceMax));
			}
			catch (err) {}
		}
		
		this.returned();
	}
	curve() {
		this.explode();
	}
	hit_event() {
		this.explode();
	}
}

class BoomerangGravity extends Boomerang {
	speed = 0.35;
	damage = 0;
	go() {
		const nearby_list = this.rang.dimension.getEntities({ location: { x: this.rang.location.x, y: this.rang.location.y, z: this.rang.location.z }, minDistance: 0.2, maxDistance: 10, excludeFamilies: [ "inanimate" ] });
		for (const entity of nearby_list)
		{
			if (entity.id != this.owner.id)
			{
				let distance = Math.pow(Math.pow(this.rang.location.x - entity.location.x, 2) + Math.pow(this.rang.location.z - entity.location.z, 2),0.5);
				try {
					entity.applyKnockback( util.clamp(this.rang.location.x - entity.location.x, -5, 5) / 5, util.clamp(this.rang.location.z - entity.location.z, -5, 5) / 5, (10 - Math.min(distance, 9.9)) / 10, util.clamp(this.rang.location.y - entity.location.y, -9.9, 9.9) / 10 );
				}
				catch (err) { }
			}
		}
		super.go();
	}
	hit_event() {}
}

// Clean up active boomerangs that have been unloaded
async function boomerang_cleanup_tick() {
	let rangList = util.get_entities("all", {families: [ "ljw_boomerang" ]});
	for (const boomerang of rangList.filter((item,index) => rangList.indexOf(item) === index))
	{
		if (rang_roster[boomerang.id] == null)
		{
			try {
				world.getEntity(boomerang.getDynamicProperty("ljw_br:owner_id")).getComponent("inventory").container.addItem(new ItemStack(identify_rang(boomerang)));
				boomerang.triggerEvent("ljw_br:event_despawn");
			}
			catch (err)
			{
				boomerang.dimension.spawnItem(new ItemStack(identify_rang(boomerang)), boomerang.location);
				boomerang.triggerEvent("ljw_br:event_despawn");
			}
		}
	}
}
system.runInterval(boomerang_cleanup_tick, 100);

system.run(boomerang_cleanup_tick);