import {system, world, ItemStack} from "@minecraft/server";
import { ActionFormData, MessageFormData } from "@minecraft/server-ui";
import * as debug from "../debug";

const boomerangList = [
	{name: "Wooden Boomerang", category: "Standard", description: "A common, trusty boomerang made of wood.", ingredients: "1 leather + 2 Stick", iconPath: "textures/ljw/br/items/boomerang_wooden"},
	{name: "Stone Boomerang", category: "Standard", description: "A sturdy, trusty boomerang made of stone.", ingredients: "1 leather + 2 Cobblestone", iconPath: "textures/ljw/br/items/boomerang_stone"}, 
	{name: "Iron Boomerang", category: "Standard", description: "A dependable, trusty boomerang made of iron.", ingredients: "1 leather + 2 Iron Ingot", iconPath: "textures/ljw/br/items/boomerang_iron"}, 
	{name: "Golden Boomerang", category: "Standard", description: "A fancy, trusty boomerang made of gold.", ingredients: "1 Leather + 2 Gold Ingot", iconPath: "textures/ljw/br/items/boomerang_golden"}, 
	{name: "Diamond Boomerang", category: "Standard", description: "A strong, trusty boomerang made of diamond.", ingredients: "1 Leather + 2 Diamond", iconPath: "textures/ljw/br/items/boomerang_diamond"}, 
	{name: "Netherite Boomerang", category: "Standard", description: "A powerful, trusty boomerang made of netherite.", ingredients: "NETHERITE-SMITHING", iconPath: "textures/ljw/br/items/boomerang_netherite"}, 
	{name: "Crimson Boomerang", category: "Standard", description: "Crafted from the wild elements of Crimson Forests, this boomerang can pierce walls and mobs.", ingredients: "1 Leather + 2 Stick + 1 Crimson Fungus", iconPath: "textures/ljw/br/items/boomerang_crimson"}, 
	{name: "Warped Boomerang", category: "Standard", description: "Crafted from the twisted elements of Warped Forests, this boomerang slowly pushes mobs away.", ingredients: "1 Leather + 2 Stick + 1 Warped Fungus", iconPath: "textures/ljw/br/items/boomerang_warped"}, 
	{name: "Gilded Boomerang", category: "Standard", description: "An ancient Nether design, this fancy boomerang flies in a double loop when thrown correctly.", ingredients: "1 Blackstone + 2 Gilded Blackstone", iconPath: "textures/ljw/br/items/boomerang_gilded"}, 
	{name: "Minerang", category: "Utility", description: "This boomerang is tough enough to cut through solid stone as it flies.", ingredients: "1 Iron Boomerang + 1 Iron Pickaxe", iconPath: "textures/ljw/br/items/boomerang_minerang"}, 
	{name: "Cutterang", category: "Utility", description: "This boomerang chops through even the toughest foliage from any dimension.", ingredients: "1 Iron Boomerang + 1 Shears", iconPath: "textures/ljw/br/items/boomerang_cutterang"}, 
	{name: "Enderang", category: "Utility", description: "Instead of returning, this boomerang brings its user to itself!", ingredients: "1 Ender Pearl + 2 Popped Chorus Fruit", iconPath: "textures/ljw/br/items/boomerang_enderang"}, 
	{name: "Hopperang", category: "Utility", description: "A boomerang that picks up items on its way back to its owner.", ingredients: "1 Iron Boomerang + 1 Hopper", iconPath: "textures/ljw/br/items/boomerang_hopperang"}, 
	{name: "Gatherang", category: "Utility", description: "A boomerang that searches for items that its owner is already carrying.", ingredients: "1 Hopperang + 1 Amethyst Shard", iconPath: "textures/ljw/br/items/boomerang_gatherang"}, 
	{name: "Collectorang", category: "Utility", description: "A boomerang that collects all nearby items and brings them back to its owner.", ingredients: "1 Hopperang + 1 Eye of Ender", iconPath: "textures/ljw/br/items/boomerang_collectorang"}, 
	{name: "Frost Boomerang", category: "Special", description: "A boomerang cold enough to slow enemies. Creates frosted ice if it hits water.", ingredients: "1 Blue Ice + 2 Packed Ice", iconPath: "textures/ljw/br/items/boomerang_frost"}, 
	{name: "Magmarang", category: "Special", description: "A boomerang that carries the power of fire. Will extinguish if it contacts water. Magma, fire or lava can reignite it.", ingredients: "1 Blackstone + 2 Magma", iconPath: "textures/ljw/br/items/boomerang_magma_charged"}, 
	{name: "Hydrorang", category: "Special", description: "An ancient boomerang made of prismarine that slices through water.", ingredients: "1 Dark Prismarine + 2 Prismarine Shard", iconPath: "textures/ljw/br/items/boomerang_hydrorang"}, 
	{name: "Echorang", category: "Special", description: "A mysterious boomerang that can be controlled remotely...", ingredients: "1 Echo Shard + 2 Cobbled Deepslate", iconPath: "textures/ljw/br/items/boomerang_echorang"}, 
	{name: "Enigmarang", category: "Special", description: "Will it push enemies away? Will it explode? You never know how this boomerang will behave when it flies!", ingredients: "1 Iron Boomerang + 1 Bottle o' Enchanting + 1 Lapis Lazuli", iconPath: "textures/ljw/br/items/boomerang_enigmarang"}, 
	{name: "Slime Boomerang", category: "Special", description: "While it doesn't do any damage, this boomerang is really good at pushing mobs around.", ingredients: "8 Slime Ball", iconPath: "textures/ljw/br/items/boomerang_slime"}, 
	{name: "TNT Boomerang", category: "Special", description: "Despite its name, this boomerang doesn't actually destroy blocks. Explodes on contact, or after a short while.", ingredients: "1 Iron Boomerang + 1 TNT", iconPath: "textures/ljw/br/items/boomerang_tnt"}, 
	{name: "Turborang", category: "Special", description: "The fastest, farthest flying boomerang. Does more damage the further it flies!", ingredients: "1 Iron Boomerang + 1 Firework Rocket", iconPath: "textures/ljw/br/items/boomerang_turborang"}, 
	{name: "Void Boomerang", category: "Special", description: "Give this boomerang a swing to make a limitless number of copies!", ingredients: "1 Popped Chorus Fruit + 2 End Stone", iconPath: "textures/ljw/br/items/boomerang_void"},
	{name: "Peppermint Boomerang", category: "Special", description: "A festive boomerang that gives its owner a sweet speed boost when attacking mobs.", ingredients: "1 Red Dye + 1 White Dye + 2 Sugar", iconPath: "textures/ljw/br/items/boomerang_peppermint"},
	{name: "Snow Boomerang", category: "Special", description: "A mostly harmless boomerang made of snow.", ingredients: "3 Snowballs", iconPath: "textures/ljw/br/items/boomerang_snow"},
	{name: "Applerang", category: "Special", description: "This boomerang causes airborne mobs to come crashing down.", ingredients: "3 Apples", iconPath: "textures/ljw/br/items/boomerang_applerang"},
	{name: "Recoverang", category: "Special", description: "Instead of dealing damage, this boomerang heals mobs that it hits!", ingredients: "1 Iron Boomerang + 1 Golden Apple", iconPath: "textures/ljw/br/items/boomerang_recoverang"},
	{name: "Wolfwind", category: "Unique", description: "A fierce boomerang that targets all nearby monsters before returning to its owner.", ingredients: "1 Iron Boomerang + 1 Bone", iconPath: "textures/ljw/br/items/boomerang_wolfwind"}, 
	{name: "Breezyrang", category: "Unique", description: "This boomerang explodes in a gust of wind!", ingredients: "1 Tuff + 1 Breeze Rod + 2 Oxidized Copper", iconPath: "textures/ljw/br/items/boomerang_breezyrang"},
	{name: "Raptorang", category: "Unique", description: "This boomerang has a very short flight path and can hit multiple mobs.", ingredients: "1 Wooden Boomerang + 1 Bone Block", iconPath: "textures/ljw/br/items/boomerang_raptorang"},
	{name: "Heavenlyrang", category: "Unique", description: "This boomerang bestows positive effects on mobs that it hits.", ingredients: "1 Iron Boomerang + 1 Golden Boomerang + 1 Golden Apple", iconPath: "textures/ljw/br/items/boomerang_heavenlyrang"},
	{name: "Jellyrang", category: "Unique", description: "This glowing boomerang can travel through air and water!", ingredients: "1 Slime Boomerang + 1 Prismarine Boomerang", iconPath: "textures/ljw/br/items/boomerang_jellyrang"},
	{name: "Torcharang", category: "Unique", description: "Wait.. This boomerang doesn't even glow?!", ingredients: "1 Wooden Boomerang + 1 Torchflower", iconPath: "textures/ljw/br/items/boomerang_torcharang"},
	{name: "Boomerangs Guide", category: "Informational", description: "The guide that you're currently reading can be thrown if you use it while sneaking! Doesn't do much damage, but it can be handy in a pinch.", ingredients: "INFO-BOOK", iconPath: "textures/ljw/br/items/info_book"}
]
const classColors = {
	"Standard": "8",
	"Utility": "2",
	"Special": "5",
	"Unique": "n",
	"Informational": "s"
}
const specialTableCraftingPattern = `\n\n[1][2][ ]\n[3][3][ ]\n[3][3][ ]\n\n1 = Stick\n2 = Leather\n3 = Planks\n\nIt's like making a smithing table, except you use sticks & leather instead of iron.`;

world.afterEvents.playerSpawn.subscribe(e => {
	const player = e.player;
		
	if (!(player.getDynamicProperty("ljw_br:info_book_init") ?? false))
	{
		give_info(player);
		player.setDynamicProperty("ljw_br:info_book_init", true);
	}
});

function give_info(player) {
	player.dimension.spawnItem(new ItemStack("ljw_br:info_book"), player.location);
}

system.afterEvents.scriptEventReceive.subscribe(e => {
    const entity = e.sourceEntity;
    if (e.id == "ljw_br:info_book") {
		if (e.message == "get")
		{
			if (entity == null)
			{
				debug.error("Error: 'ljw_br:info_book get' must be run from an entity.");
				return;
			}
			
			give_info(entity);
		}
		else if (e.message == "reset")
		{
			for (const player of world.getPlayers())
				player.setDynamicProperty("ljw_br:info_book_init", false);
		}
    }
});

world.afterEvents.itemUse.subscribe(e => {
	const target = e.source;
	if ((e.itemStack.typeId == "ljw_br:info_book") && !target.isSneaking)
		open_guide(target);
});

function open_guide(player) {
	try {
		const guideDialog = new ActionFormData();
		const welcomeGreeting = "                 §aWelcome!§r";
		const welcomeMessage = "This guide contains information about the §aBoomerangs§r Add-On.\n\nSelect a category.";
		const divider = "_".repeat(32);
		
		guideDialog.title("Boomerangs Guide");
		guideDialog.body(`${welcomeGreeting}\n${welcomeMessage}\n${divider}`);
		
		guideDialog.button("§eGetting Started§8\n(Important Info!)", "textures/blocks/sapling_oak");
		guideDialog.button("Boomerangs", boomerangList[Math.floor(Math.random() * boomerangList.length)].iconPath);
		guideDialog.button("About", "textures/ljw/br/items/info_book");
		
		guideDialog.show(player).then(response => {
			if (response.canceled || response.selection == undefined)
				return;
			else if (response.selection == 0)
				open_initial_steps(player);
			else if (response.selection == 1)
				open_boomerang_list(player);
			else if (response.selection == 2)
				open_credits(player);
		});
	}
	catch (err) {
		debug.error(`Info Book error: ${err}`)
	}
}

function open_initial_steps(player) {
	try {
		const guideDialog = new ActionFormData();
		const welcomeMessage = "Select a topic to learn more about.";
		const divider = "_".repeat(32);
		
		guideDialog.title("Getting Started");
		guideDialog.body(`In order to craft most of the items in this Add-On, you'll need a §aBoomerang Crafting Table§r.\n\nTo craft the table, follow this pattern:${specialTableCraftingPattern}`);
		
		guideDialog.button("Back");
		guideDialog.button("Exit");
		
		guideDialog.show(player).then(response => {
			if (response.canceled || response.selection == undefined)
				return;
			else if (response.selection == 0)
				open_guide(player);
			else if (response.selection == 1)
				return;
		});
	}
	catch (err) {
		debug.error(`Info Book error: ${err}`)
	}
}

function open_boomerang_list(player) {
	try {
		const guideDialog = new ActionFormData();
		const welcomeMessage = "Select a boomerang to learn more about.";
		const divider = "_".repeat(32);
		
		guideDialog.title("Boomerangs");
		guideDialog.body(`${welcomeMessage}\n${divider}`);
		
		for (const boomerang of boomerangList)
			guideDialog.button(`§${classColors[boomerang.category]}${boomerang.name}`, boomerang.iconPath);
		
		guideDialog.show(player).then(response => {
			if (response.canceled || response.selection == undefined)
				open_guide(player);
			else
				explain_item(player, boomerangList[response.selection]);
		});
	}
	catch (err) {
		debug.error(`Info Book error: ${err}`)
	}
}

function explain_item(player, itemEntry) {
	try {
		const guideDialog = new ActionFormData();
		const how_to_obtain = itemEntry.ingredients == "CRAFTING-TABLE" ? `Crafting:\n${specialTableCraftingPattern}` : itemEntry.ingredients == "NOT-CRAFTABLE" ? `Can't be crafted.` : itemEntry.ingredients == "NETHERITE-SMITHING" ? `Crafted via Netherite Upgrade at Smithing Table.` : itemEntry.ingredients == "INFO-BOOK" ? `Crafting: 1 Stick + 1 Leather + 1 Book` : `Crafting:\n${itemEntry.ingredients} @ Boomerang Crafting Table`
		const description = `Class: §${classColors[itemEntry.category]}${itemEntry.category}§r\n\n${how_to_obtain}\n\n§o${itemEntry.description}§r`;
		
		guideDialog.title(itemEntry.name);
		guideDialog.body(`${description}\n\n`);
		
		guideDialog.button("Back");
		guideDialog.button("Exit");
		
		guideDialog.show(player).then(response => {
			if (response.canceled || response.selection == undefined)
				return;
			else if (response.selection == 0)
				open_guide(player);
			else if (response.selection == 1)
				return;
		});
	}
	catch (err) {
		debug.error(`Info Book error: ${err}`)
	}
}

function open_credits(player) {
	try {
		const guideDialog = new ActionFormData();
		
		guideDialog.title("About");
		guideDialog.body(`§lBoomerangs§r v1.0.0\nCreated by JWolf\n\nFor questions or support, please reach out to me on X (formerly known as Twitter): §b@JWolf1765§r\n\n`);
		
		guideDialog.button("Changelog");
		guideDialog.button("Special Thanks");
		guideDialog.button("Back");
		guideDialog.button("Exit");
		
		guideDialog.show(player).then(response => {
			if (response.canceled || response.selection == undefined)
				return;
			else if (response.selection == 0)
				open_changelog(player);
			else if (response.selection == 1)
				open_thanks(player);
			else if (response.selection == 2)
				open_guide(player);
			else if (response.selection == 3)
				return;
		});
	}
	catch (err) {
		debug.error(`Info Book error: ${err}`)
	}
}

function open_changelog(player) {
	try {
		const guideDialog = new ActionFormData();
		let content = "";
		
		content += `§lBoomerangs§r v1.0.0\n- Initial Release!`
		content += `\n\n`
		
		guideDialog.title("Changelog");
		guideDialog.body(content);
		
		guideDialog.button("Back");
		guideDialog.button("Exit");
		
		guideDialog.show(player).then(response => {
			if (response.canceled || response.selection == undefined)
				return;
			else if (response.selection == 0)
				open_guide(player);
			else if (response.selection == 1)
				return;
		});
	}
	catch (err) {
		debug.error(`Info Book error: ${err}`)
	}
}

const thanksList = [
	{ name: "Beez", social: "@itisbeez", role: "Art", contributions: [ "Boomerang Item Art Overhaul", "Boomerang Crafting Table Art" ]}
]

const roleIconMappings = {
	"Art": "textures/ui/color_picker",
	"Audio": "textures/items/record_cat",
	"Testing": "textures/ui/controller_glyph_color"
}

function open_thanks(player) {
	try {
		const guideDialog = new ActionFormData();
		let content = "";
		
		content += `Thanks to everyone who contributed to this project!`
		content += `\n\n`
		
		guideDialog.title("Special Thanks");
		guideDialog.body(content);
		
		for (const entry of thanksList) {
			guideDialog.button(`${entry.name}`, roleIconMappings[entry.role]);
		}
		
		guideDialog.button("Back");
		guideDialog.button("Exit");
		
		guideDialog.show(player).then(response => {
			if (response.canceled || response.selection == undefined)
				return;
			else if (response.selection == thanksList.length)
				open_credits(player);
			else if (response.selection == thanksList.length + 1)
				return;
			else
				explain_thanks(player, response.selection);
		});
	}
	catch (err) {
		debug.error(`Info Book error: ${err}`)
	}
}

function explain_thanks(player, num) {
	try {
		const guideDialog = new ActionFormData();
		let content = "";
		
		content += `§b${thanksList[num].social}§r`;
		content += `\n`;
		content += `Role: ${thanksList[num].role}`;
		content += `\n\n`;
		content += `Contributions:`;
		content += `\n`;
		
		for (const entry of thanksList[num].contributions) {
			content += `- ${entry}`;
			content += `\n`;
		}
		content += `\n\n`;
		
		guideDialog.title(thanksList[num].name);
		guideDialog.body(content);
		
		guideDialog.button("Back");
		guideDialog.button("Exit");
		
		guideDialog.show(player).then(response => {
			if (response.canceled || response.selection == undefined)
				return;
			else if (response.selection == 0)
				open_thanks(player);
			else if (response.selection == 3)
				return;
		});
	}
	catch (err) {
		debug.error(`Info Book error: ${err}`)
	}
}